%{
  
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <malloc.h>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include "parser.hh"


  extern "C"
  {
    int yyparse(void);
    int yylex(void); 
    int yywrap()
    {
      return 1;
    }
    extern int lineno;
#include "../pnode/pnode.h"
  }
  
  void yyerror(const char *str)
    {
      fprintf(stderr,"error: %s at line %d\n",str,lineno);
    }
  
  char tmpstr[250];

  extern map< string, pnode_ptr> evcond;
  extern vector < map< string, pnode_ptr> > evcondlist;
  extern map< string, pnode_ptr> evalcond;
  extern vector < pnode_ptr > iscond;
  extern vector < pnode_ptr > formulaelist;
  extern map< string, vector < string > > groups;

  vector < string > names_p;
%}


%union {
  char *str;
  pnode_ptr pnode;
}

%token AGENT LSTATE ACTION PROTOCOL EV END LB RB LCB RCB IF EQ AND
%token OR NOT COLON COMMA DOT SEMICOLON EVALUATION INITSTATES GROUPS
%token FORMULAE IMPLIES AG EG AX EX AF EF A E UNTIL K GK GCK
%token <str> ID

%left LB RB
%left ID
%right IMPLIES
%left OR 
%left AND
%left NOT
%left EX AX EF AF EG AG E A UNTIL 
%left K GK GCK
%left EQ
%left DOT

%type <pnode> boolcond
%type <pnode> evaboolcond
%type <pnode> isboolcond
%type <pnode> formula

%%

is: agents evaluation istates groups formulae;

agents: agent

  | agents agent;

agent: AGENT ID agbody END AGENT {
  agtnames.push_back($2);
  //cout << "Name: " << $2 <<endl;
  //cout << "Local states: " << endl;
  //for( igs = lstates.begin(); igs != lstates.end(); igs++) {
  //  cout << *igs << endl;
  //}  
};

agbody: lstatedef actiondef protdef evdef;

/* LOCAL STATES */
lstatedef: LSTATE EQ lstatedeflist;

lstatedeflist: LCB lstateidlist RCB SEMICOLON {
  lstatelist.push_back(lstates);
  lstates.clear();
};

lstateidlist: ID { 
  /* <DEBUG> */
  lstates.push_back($1);
  //printf("Local state: %s\n",$1);
  /* </DEBUG> */
}
| lstateidlist COMMA ID {
  /* <DEBUG> */
  //printf("Local state: %s\n",$3);
  lstates.push_back($3);
  /* </DEBUG> */
};

/* ACTIONS */
actiondef: ACTION EQ actiondeflist;

actiondeflist: LCB actionidlist RCB SEMICOLON
{
  actionlist.push_back(actions);
  actions.clear();
};


actionidlist: ID {
		/* <DEBUG> */
  actions.push_back($1);
		/* </DEBUG> */
		}
	| actionidlist COMMA ID {
                /* <DEBUG> */
	  actions.push_back($3);
                /* </DEBUG> */
		};


/* PROTOCOL */
protdef: PROTOCOL COLON protdeflist END PROTOCOL 
{
  protlist.push_back(protocol);
  protocol.erase(protocol.begin(),protocol.end());
};


protdeflist: protline 
| protdeflist protline;


protline: ID COLON protenabledidlist {
  /* <DEBUG> */
  //printf("Protocol LS: %s\n",$1);
  /* </DEBUG> */
  protocol[$1] = strvect;
  strvect.clear();
};
protenabledidlist:  LCB enabledidlist RCB SEMICOLON;

enabledidlist: ID {
  /* <DEBUG> */
  //printf("Action enabled: %s\n",$1);
  strvect.push_back($1);
  /* </DEBUG> */
}
| enabledidlist COMMA ID {
  /* <DEBUG> */
  //printf("Action enabled: %s\n",$3);
  strvect.push_back($3);
  /* </DEBUG> */
};


/* EVOLUTION DEFINITION */
evdef: EV COLON evdeflist END EV
{
  evcondlist.push_back(evcond);
  evcond.erase(evcond.begin(),evcond.end());
};

evdeflist: evline | evdeflist evline;

evline: ID IF boolcond SEMICOLON {
                /* <DEBUG> */
  //			printf("Next LS STAMPA!!!: %s\n",$1);
  //printf("Next LS=%s if ",$1);
  evcond[$1]=$3;
  //  parse_tree = $3;
  //np.push_back(parse_tree);
  //		print_pnode(parse_tree);
  //		printf("\n");
                /* </DEBUG> */
		};

boolcond: LB boolcond RB { 
  $$ = $2; 
}
| boolcond AND boolcond {
  /* <DEBUG> */
  $$ = new_pnode(1,"",$1,$3);
  //  printf("AND condition\n");
  /* </DEBUG> */
}
| boolcond OR boolcond {
  /* <DEBUG> */
  $$ = new_pnode(2,"",$1,$3);
  //printf("OR condition\n");
  /* </DEBUG> */
}
| NOT boolcond {
  /* <DEBUG> */
  $$ = new_pnode(3,"",$2,NULL);
  //printf("NOT condition\n");
  /* </DEBUG> */
}
| LSTATE EQ ID {	
  /* <DEBUG> */
  
  // strcpy(tmpstr,"Lstate=");
  // sprintf(tmpstr,"%s=%s",tmpstr,$3);
  // strcat(tmpstr,$3);
  sprintf(tmpstr,"Lstate=%s",$3);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //free(tmpstr);
  //printf("LS assign: %s \n",$3);
  /* </DEBUG> */
}
| ID DOT LSTATE EQ ID {	
  /* <DEBUG> */
  sprintf(tmpstr,"%s.Lstate=%s",$1,$5);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //printf("Ext. LS assign: %s = %s \n",$1,$5);
  /* </DEBUG> */
}
| ACTION EQ ID {
  /* <DEBUG> */
  sprintf(tmpstr,"Action=%s",$3);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  /* </DEBUG> */
}
| ID DOT ACTION EQ ID {	
  /* <DEBUG> */
  sprintf(tmpstr,"%s.Action=%s",$1,$5);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //printf("Ext. Act. assign: %s = %s \n",$1,$5);
  /* </DEBUG> */
};


/* EVALUATION */
evaluation: EVALUATION evalist END EVALUATION;

evalist: evaline | evalist evaline;

evaline: ID IF evaboolcond SEMICOLON {
  /* <DEBUG> */
  evalcond[$1]=$3;
};
evaboolcond: LB evaboolcond RB { 
  $$ = $2; 
}
| evaboolcond AND evaboolcond {
  /* <DEBUG> */
  $$ = new_pnode(1,"",$1,$3);
  //  printf("AND condition\n");
  /* </DEBUG> */
}
| evaboolcond OR evaboolcond {
  /* <DEBUG> */
  $$ = new_pnode(2,"",$1,$3);
  //printf("OR condition\n");
  /* </DEBUG> */
}
| NOT evaboolcond {
  /* <DEBUG> */
  $$ = new_pnode(3,"",$2,NULL);
  //printf("NOT condition\n");
  /* </DEBUG> */
}
| LSTATE EQ ID {	
  /* <DEBUG> */
  
  // strcpy(tmpstr,"Lstate=");
  // sprintf(tmpstr,"%s=%s",tmpstr,$3);
  // strcat(tmpstr,$3);
  $$ = new_pnode(LEAF,$3,NULL,NULL);
  //free(tmpstr);
  //printf("LS assign: %s \n",$3);
  /* </DEBUG> */
}
| ID DOT LSTATE EQ ID {	
  /* <DEBUG> */
  sprintf(tmpstr,"%s.Lstate=%s",$1,$5);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //printf("Ext. LS assign: %s = %s \n",$1,$5);
  /* </DEBUG> */
}
| ACTION EQ ID {
  /* <DEBUG> */
  sprintf(tmpstr,"Action=%s",$3);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  /* </DEBUG> */
}
| ID DOT ACTION EQ ID {	
  /* <DEBUG> */
  sprintf(tmpstr,"%s.Action=%s",$1,$5);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //printf("Ext. Act. assign: %s = %s \n",$1,$5);
  /* </DEBUG> */
};



/* INITIAL STATES */


istates: INITSTATES isboolcond SEMICOLON END INITSTATES
{
  iscond.push_back($2);  
};

isboolcond: LB isboolcond RB { 
  $$ = $2; 
}
| isboolcond AND isboolcond {
  /* <DEBUG> */
  $$ = new_pnode(1,"",$1,$3);
  //  printf("AND condition\n");
  /* </DEBUG> */
}
| isboolcond OR isboolcond {
  /* <DEBUG> */
  $$ = new_pnode(2,"",$1,$3);
  //printf("OR condition\n");
  /* </DEBUG> */
}
| NOT isboolcond {
  /* <DEBUG> */
  $$ = new_pnode(3,"",$2,NULL);
  //printf("NOT condition\n");
  /* </DEBUG> */
}
| LSTATE EQ ID {	
  /* <DEBUG> */
  
  // strcpy(tmpstr,"Lstate=");
  // sprintf(tmpstr,"%s=%s",tmpstr,$3);
  // strcat(tmpstr,$3);
  $$ = new_pnode(LEAF,$3,NULL,NULL);
  //free(tmpstr);
  //printf("LS assign: %s \n",$3);
  /* </DEBUG> */
}
| ID DOT LSTATE EQ ID {	
  /* <DEBUG> */
  sprintf(tmpstr,"%s.Lstate=%s",$1,$5);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //printf("Ext. LS assign: %s = %s \n",$1,$5);
  /* </DEBUG> */
}
| ACTION EQ ID {
  /* <DEBUG> */
  sprintf(tmpstr,"Action=%s",$3);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  /* </DEBUG> */
}
| ID DOT ACTION EQ ID {	
  /* <DEBUG> */
  sprintf(tmpstr,"%s.Action=%s",$1,$5);
  $$ = new_pnode(LEAF,tmpstr,NULL,NULL);
  //printf("Ext. Act. assign: %s = %s \n",$1,$5);
  /* </DEBUG> */
};



/* Groups (not necessary) */

groups: GROUPS groupdeflist END GROUPS;

groupdeflist: groupline 
| groupdeflist groupline;

groupline: ID EQ LCB namelist RCB SEMICOLON
{
  groups[$1] = names_p;
  names_p.clear();
}
;

namelist: ID 
{
  names_p.push_back($1);
}
| namelist COMMA ID 
{
  names_p.push_back($3);
}
;



/* FORMULAE TO CHECK */


formulae: FORMULAE formlist END FORMULAE;

formlist: formula SEMICOLON { formulaelist.push_back($1); } 
| formlist formula SEMICOLON { formulaelist.push_back($2); };

formula: LB formula RB { 
  $$ = $2; 
}
| formula AND formula {
  /* <DEBUG> */
  $$ = new_pnode(1,"",$1,$3);
  //  printf("AND condition\n");
  /* </DEBUG> */
}
| formula OR formula {
  /* <DEBUG> */
  $$ = new_pnode(2,"",$1,$3);
  //printf("OR condition\n");
  /* </DEBUG> */
}
| NOT formula {
  /* <DEBUG> */
  $$ = new_pnode(3,"",$2,NULL);
  //printf("NOT condition\n");
  /* </DEBUG> */
}
| formula IMPLIES formula 
{
  $$ = new_pnode(4,"",$1,$3);
}
| AG formula 
{
  $$ = new_pnode(10,"",$2,NULL);
}
| EG formula 
{
  $$ = new_pnode(11,"",$2,NULL);
}
| AX formula 
{
  $$ = new_pnode(12,"",$2,NULL);
}
| EX formula 
{
  $$ = new_pnode(13,"",$2,NULL);
}
| AF formula 
{
  $$ = new_pnode(14,"",$2,NULL);
}
| EF formula 
{
  $$ = new_pnode(15,"",$2,NULL);
}
| A LB formula UNTIL formula RB 
{
  $$ = new_pnode(16,"",$3,$5);
}
| E LB formula UNTIL formula RB 
{
  $$ = new_pnode(17,"",$3,$5);
}
| K LB ID COMMA formula RB 
{
  $$ = new_pnode(30,$3,$5,NULL);
}
| GK LB ID COMMA formula RB
{
  // Knowledge in a group (everybody knows)
  $$ = new_pnode(31,$3,$5,NULL);
}
| GCK LB ID COMMA formula RB
{
  // Common knowledge in a group
  $$ = new_pnode(32,$3,$5,NULL);
}
| ID {	
  $$ = new_pnode(LEAF,$1,NULL,NULL);
};

%%
